#!/usr/bin/env python

class InvalidCallbackResult(Exception):
    pass


class InvalidFuncType(Exception):
    pass


class InvalidRequestMethod(Exception):
    pass


class NotImplementedParseError(Exception):
    pass


class NothingMatchedError(Exception):
    pass
